from c4v.scraper.scraper    import bulk_scrape, scrape
from c4v.scraper.settings   import INSTALLED_CRAWLERS, SUPPORTED_DOMAINS, INSTALLED_SCRAPERS
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData
